@@ -1,7 +1,7 @@
 /* encrypt.c - Main encryption driver
  * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
  *               2006, 2009 Free Software Foundation, Inc.
- * Copyright (C) 2016 g10 Code GmbH
+ * Copyright (C) 2016, 2023 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -17,6 +17,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include <config.h>
@@ -45,6 +46,30 @@ static int encrypt_simple( const char *filename, int mode, int use_seskey );
 static int write_pubkey_enc_from_list (ctrl_t ctrl,
                                        PK_LIST pk_list, DEK *dek, iobuf_t out);
 
+
+/* Helper for show the "encrypted for USER" during encryption.
+ * PUBKEY_USAGE is used to figure out whether this is an ADSK key.  */
+static void
+show_encrypted_for_user_info (ctrl_t ctrl, unsigned int pubkey_usage,
+                              PKT_pubkey_enc *enc, DEK *dek)
+{
+  char *ustr = get_user_id_string_native (ctrl, enc->keyid);
+  if ((pubkey_usage & PUBKEY_USAGE_RENC))
+    {
+      char *tmpustr = xstrconcat (ustr, " [ADSK]", NULL);
+      xfree (ustr);
+      ustr = tmpustr;
+    }
+  log_info (_("%s/%s.%s encrypted for: \"%s\"\n"),
+            openpgp_pk_algo_name (enc->pubkey_algo),
+            openpgp_cipher_algo_name (dek->algo),
+            dek->use_aead? openpgp_aead_algo_name (dek->use_aead)
+            /**/         : "CFB",
+            ustr );
+  xfree (ustr);
+}
+
+
 /****************
  * Encrypt FILENAME with only the symmetric cipher.  Take input from
  * stdin if FILENAME is NULL.  If --force-aead is used we use an SKESK.
@@ -67,6 +92,148 @@ encrypt_store (const char *filename)
 }
 
 
+/* Create and setup a DEK structure and print appropriate warnings.
+ * PK_LIST gives the list of public keys.  Always returns a DEK.  The
+ * actual session needs to be added later.  */
+static DEK *
+create_dek_with_warnings (pk_list_t pk_list)
+{
+  DEK *dek;
+
+  dek = xmalloc_secure_clear (sizeof *dek);
+  if (!opt.def_cipher_algo)
+    {
+      /* Try to get it from the prefs.  */
+      dek->algo = select_algo_from_prefs (pk_list, PREFTYPE_SYM, -1, NULL);
+      if (dek->algo == -1)
+        {
+          /* If does not make sense to fallback to the rfc4880
+           * required 3DES if we will reject that algo later.  Thus we
+           * fallback to AES anticipating RFC4880bis rules.  */
+          if (opt.flags.allow_old_cipher_algos)
+            dek->algo = CIPHER_ALGO_3DES;
+          else
+            dek->algo = CIPHER_ALGO_AES;
+        }
+
+      /* In case 3DES has been selected, print a warning if any key
+       * does not have a preference for AES.  This should help to
+       * identify why encrypting to several recipients falls back to
+       * 3DES. */
+      if (opt.verbose && dek->algo == CIPHER_ALGO_3DES)
+        warn_missing_aes_from_pklist (pk_list);
+    }
+  else
+    {
+      if (!opt.expert
+          && (select_algo_from_prefs (pk_list, PREFTYPE_SYM,
+                                      opt.def_cipher_algo, NULL)
+              != opt.def_cipher_algo))
+        {
+          log_info(_("WARNING: forcing symmetric cipher %s (%d)"
+                     " violates recipient preferences\n"),
+                   openpgp_cipher_algo_name (opt.def_cipher_algo),
+                   opt.def_cipher_algo);
+        }
+
+      dek->algo = opt.def_cipher_algo;
+    }
+
+  return dek;
+}
+
+
+/* Check whether all encryption keys are compliant with the current
+ * mode and issue respective status lines.  DEK has the info about the
+ * session key and PK_LIST the list of public keys.  */
+static gpg_error_t
+check_encryption_compliance (DEK *dek, pk_list_t pk_list)
+{
+  gpg_error_t err = 0;
+  pk_list_t pkr;
+  int compliant;
+
+  /* First check whether we should use the algo at all.  */
+  if (openpgp_cipher_blocklen (dek->algo) < 16
+      && !opt.flags.allow_old_cipher_algos)
+    {
+      log_error (_("cipher algorithm '%s' may not be used for encryption\n"),
+		 openpgp_cipher_algo_name (dek->algo));
+      if (!opt.quiet)
+        log_info (_("(use option \"%s\" to override)\n"),
+                  "--allow-old-cipher-algos");
+      err = gpg_error (GPG_ERR_CIPHER_ALGO);
+      goto leave;
+    }
+
+  /* Now check the compliance.  */
+  if (! gnupg_cipher_is_allowed (opt.compliance, 1, dek->algo,
+                                 GCRY_CIPHER_MODE_CFB))
+    {
+      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
+		 openpgp_cipher_algo_name (dek->algo),
+		 gnupg_compliance_option_string (opt.compliance));
+      err = gpg_error (GPG_ERR_CIPHER_ALGO);
+      goto leave;
+    }
+
+  if (!gnupg_rng_is_compliant (opt.compliance))
+    {
+      err = gpg_error (GPG_ERR_FORBIDDEN);
+      log_error (_("%s is not compliant with %s mode\n"),
+                 "RNG",
+                 gnupg_compliance_option_string (opt.compliance));
+      write_status_error ("random-compliance", err);
+      goto leave;
+    }
+
+  /* From here on we only test for CO_DE_VS - if we ever want to
+   * return other compliance mode values we need to change this to
+   * loop over all those values.  */
+  compliant = gnupg_gcrypt_is_compliant (CO_DE_VS);
+
+  if (!gnupg_cipher_is_compliant (CO_DE_VS, dek->algo, GCRY_CIPHER_MODE_CFB))
+    compliant = 0;
+
+  for (pkr = pk_list; pkr; pkr = pkr->next)
+    {
+      PKT_public_key *pk = pkr->pk;
+      unsigned int nbits = nbits_from_pk (pk);
+
+      if (!gnupg_pk_is_compliant (opt.compliance, pk->pubkey_algo, 0,
+                                  pk->pkey, nbits, NULL))
+        log_info (_("WARNING: key %s is not suitable for encryption"
+                    " in %s mode\n"),
+                  keystr_from_pk (pk),
+                  gnupg_compliance_option_string (opt.compliance));
+
+      if (compliant
+          && !gnupg_pk_is_compliant (CO_DE_VS, pk->pubkey_algo, 0, pk->pkey,
+                                     nbits, NULL))
+        compliant = 0; /* Not compliant - reset flag.  */
+    }
+
+  /* If we are compliant print the status for de-vs compliance.  */
+  if (compliant)
+    write_status_strings (STATUS_ENCRYPTION_COMPLIANCE_MODE,
+                          gnupg_status_compliance_flag (CO_DE_VS),
+                          NULL);
+
+  /* Check whether we should fail the operation.  */
+  if (opt.flags.require_compliance
+      && opt.compliance == CO_DE_VS
+      && !compliant)
+    {
+      compliance_failure ();
+      err = gpg_error (GPG_ERR_FORBIDDEN);
+      goto leave;
+    }
+
+ leave:
+  return err;
+}
+
+
 /* Encrypt a session key using DEK and store a pointer to the result
  * at R_ENCKEY and its length at R_ENCKEYLEN.
  *
@@ -77,7 +244,7 @@ encrypt_store (const char *filename)
  * stored at R_SESKEY.  If AEAD_ALGO is not 0 the given AEAD algorithm
  * is used for encryption.
  */
-gpg_error_t
+static gpg_error_t
 encrypt_seskey (DEK *dek, aead_algo_t aead_algo,
                 DEK **r_seskey, void **r_enckey, size_t *r_enckeylen)
 {
@@ -202,14 +369,6 @@ use_aead (pk_list_t pk_list, int algo)
 {
   int can_use;
 
-  if (!opt.flags.rfc4880bis)
-    {
-      if (opt.force_aead)
-        log_info ("Warning: Option %s currently requires option '%s'\n",
-                  "--force-aead", "--rfc4880bis");
-      return 0;
-    }
-
   can_use = openpgp_cipher_get_algo_blklen (algo) == 16;
 
   /* With --force-aead we want AEAD.  */
@@ -217,11 +376,11 @@ use_aead (pk_list_t pk_list, int algo)
     {
       if (!can_use)
         {
-          log_info ("Warning: request to use AEAD ignored for cipher '%s'\n",
+          log_info ("Warning: request to use OCB ignored for cipher '%s'\n",
                     openpgp_cipher_algo_name (algo));
           return 0;
         }
-      return default_aead_algo ();
+      return AEAD_ALGO_OCB;
     }
 
   /* AEAD does only work with 128 bit cipher blocklength.  */
@@ -254,10 +413,11 @@ use_mdc (pk_list_t pk_list,int algo)
 }
 
 
-/* We don't want to use use_seskey yet because older gnupg versions
-   can't handle it, and there isn't really any point unless we're
-   making a message that can be decrypted by a public key or
-   passphrase. */
+/* This function handles the --symmetric only (MODE true) and --store
+ * (MODE false) cases.  We don't want to use USE_SESKEY by default
+ * very old gnupg versions can't handle it, and there isn't really any
+ * point unless we're making a message that can be decrypted by a
+ * public key or passphrase.  */
 static int
 encrypt_simple (const char *filename, int mode, int use_seskey)
 {
@@ -336,8 +496,7 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
       if (use_seskey && s2k->mode != 1 && s2k->mode != 3)
         {
           use_seskey = 0;
-          log_info (_("can't use a symmetric ESK packet "
-                      "due to the S2K mode\n"));
+          log_info (_("can't use a SKESK packet due to the S2K mode\n"));
         }
 
       /* See whether we want to use AEAD.  */
@@ -373,17 +532,8 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
                  /**/             : "CFB");
     }
 
-  if (do_compress
-      && cfx.dek
-      && (cfx.dek->use_mdc || cfx.dek->use_aead)
-      && is_file_compressed(filename, &rc))
-    {
-      if (opt.verbose)
-        log_info(_("'%s' already compressed\n"), filename);
-      do_compress = 0;
-    }
-
-  if ( rc || (rc = open_outfile (-1, filename, opt.armor? 1:0, 0, &out )))
+  if (rc || (rc = open_outfile (GNUPG_INVALID_FD, filename, opt.armor? 1:0,
+                                0, &out )))
     {
       iobuf_cancel (inp);
       xfree (cfx.dek);
@@ -435,12 +585,12 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
 
   if ( !iobuf_is_pipe_filename (filename) && *filename && !opt.textmode )
     {
-      off_t tmpsize;
-      int overflow;
+      uint64_t tmpsize;
 
-      if ( !(tmpsize = iobuf_get_filelength(inp, &overflow))
-           && !overflow && opt.verbose)
+      tmpsize = iobuf_get_filelength(inp);
+      if (!tmpsize && opt.verbose)
         log_info(_("WARNING: '%s' is an empty file\n"), filename );
+
       /* We can't encode the length of very large files because
          OpenPGP uses only 32 bit for file sizes.  So if the
          size of a file is larger than 2^32 minus some bytes for
@@ -453,6 +603,24 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
   else
     filesize = opt.set_filesize ? opt.set_filesize : 0; /* stdin */
 
+  /* Register the cipher filter. */
+  if (mode)
+    iobuf_push_filter (out,
+                       cfx.dek->use_aead? cipher_filter_aead
+                       /**/             : cipher_filter_cfb,
+                       &cfx );
+
+  if (do_compress
+      && cfx.dek
+      && (cfx.dek->use_mdc || cfx.dek->use_aead)
+      && !opt.explicit_compress_option
+      && is_file_compressed (inp))
+    {
+      if (opt.verbose)
+        log_info(_("'%s' already compressed\n"), filename? filename: "[stdin]");
+      do_compress = 0;
+    }
+
   if (!opt.no_literal)
     {
       /* Note that PT has been initialized above in !no_literal mode.  */
@@ -472,13 +640,6 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
       pkt.pkt.generic = NULL;
     }
 
-  /* Register the cipher filter. */
-  if (mode)
-    iobuf_push_filter (out,
-                       cfx.dek->use_aead? cipher_filter_aead
-                       /**/             : cipher_filter_cfb,
-                       &cfx );
-
   /* Register the compress filter. */
   if ( do_compress )
     {
@@ -497,15 +658,13 @@ encrypt_simple (const char *filename, int mode, int use_seskey)
     {
       /* User requested not to create a literal packet, so we copy the
          plain data.  */
-    byte copy_buffer[4096];
-    int  bytes_copied;
-    while ((bytes_copied = iobuf_read(inp, copy_buffer, 4096)) != -1)
-      if ( (rc=iobuf_write(out, copy_buffer, bytes_copied)) ) {
-        log_error ("copying input to output failed: %s\n",
-                   gpg_strerror (rc) );
-        break;
-      }
-    wipememory (copy_buffer, 4096); /* burn buffer */
+      iobuf_copy (out, inp);
+      if ((rc = iobuf_error (inp)))
+        log_error (_("error reading '%s': %s\n"),
+                   iobuf_get_fname_nonnull (inp), gpg_strerror (rc));
+      else if ((rc = iobuf_error (out)))
+        log_error (_("error writing '%s': %s\n"),
+                   iobuf_get_fname_nonnull (out), gpg_strerror (rc));
     }
 
   /* Finish the stuff.  */
@@ -628,15 +787,15 @@ write_symkey_enc (STRING2KEY *symkey_s2k, aead_algo_t aead_algo,
  * Encrypt the file with the given userids (or ask if none is
  * supplied).  Either FILENAME or FILEFD must be given, but not both.
  * The caller may provide a checked list of public keys in
- * PROVIDED_PKS; if not the function builds a list of keys on its own.
+ * PROVIDED_KEYS; if not the function builds a list of keys on its own.
  *
- * Note that FILEFD is currently only used by cmd_encrypt in the
- * not yet finished server.c.
+ * Note that FILEFD and OUTPUTFD are currently only used by
+ * cmd_encrypt in the not yet finished server.c.
  */
 int
-encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
+encrypt_crypt (ctrl_t ctrl, gnupg_fd_t filefd, const char *filename,
                strlist_t remusr, int use_symkey, pk_list_t provided_keys,
-               int outputfd)
+               gnupg_fd_t outputfd)
 {
   iobuf_t inp = NULL;
   iobuf_t out = NULL;
@@ -644,7 +803,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   PKT_plaintext *pt = NULL;
   DEK *symkey_dek = NULL;
   STRING2KEY *symkey_s2k = NULL;
-  int rc = 0, rc2 = 0;
+  int rc = 0;
   u32 filesize;
   cipher_filter_context_t cfx;
   armor_filter_context_t *afx = NULL;
@@ -653,9 +812,8 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   progress_filter_context_t *pfx;
   PK_LIST pk_list;
   int do_compress;
-  int compliant;
 
-  if (filefd != -1 && filename)
+  if (filefd != GNUPG_INVALID_FD && filename)
     return gpg_error (GPG_ERR_INV_ARG);  /* Both given.  */
 
   do_compress = !!opt.compress_algo;
@@ -686,7 +844,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
 
   /* Prepare iobufs. */
 #ifdef HAVE_W32_SYSTEM
-  if (filefd == -1)
+  if (filefd == GNUPG_INVALID_FD)
     inp = iobuf_open (filename);
   else
     {
@@ -697,7 +855,7 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   if (filefd == GNUPG_INVALID_FD)
     inp = iobuf_open (filename);
   else
-    inp = iobuf_fdopen_nc (FD2INT(filefd), "rb");
+    inp = iobuf_fdopen_nc (filefd, "rb");
 #endif
   if (inp)
     iobuf_ioctl (inp, IOBUF_IOCTL_NO_CACHE, 1, NULL);
@@ -712,8 +870,8 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
       char xname[64];
 
       rc = gpg_error_from_syserror ();
-      if (filefd != -1)
-        snprintf (xname, sizeof xname, "[fd %d]", filefd);
+      if (filefd != GNUPG_INVALID_FD)
+        snprintf (xname, sizeof xname, "[fd %d]", FD_DBG (filefd));
       else if (!filename)
         strcpy (xname, "[stdin]");
       else
@@ -742,139 +900,16 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
     }
 
   /* Create a session key. */
-  cfx.dek = xmalloc_secure_clear (sizeof *cfx.dek);
-  if (!opt.def_cipher_algo)
-    {
-      /* Try to get it from the prefs.  */
-      cfx.dek->algo = select_algo_from_prefs (pk_list, PREFTYPE_SYM, -1, NULL);
-      /* The only way select_algo_from_prefs can fail here is when
-         mixing v3 and v4 keys, as v4 keys have an implicit preference
-         entry for 3DES, and the pk_list cannot be empty.  In this
-         case, use 3DES anyway as it's the safest choice - perhaps the
-         v3 key is being used in an OpenPGP implementation and we know
-         that the implementation behind any v4 key can handle 3DES.
-         Note that we do not support v3 keys since version 2.2 so the
-         above description gives only historical background. */
-      if (cfx.dek->algo == -1)
-        {
-          /* If does not make sense to fallback to the rfc4880
-           * required 3DES if we will reject that algo later.  Thus we
-           * fallback to AES anticipating RFC4880bis rules.  */
-          if (opt.flags.allow_old_cipher_algos)
-            cfx.dek->algo = CIPHER_ALGO_3DES;
-          else
-            cfx.dek->algo = CIPHER_ALGO_AES;
-        }
+  cfx.dek = create_dek_with_warnings (pk_list);
 
-      /* In case 3DES has been selected, print a warning if any key
-         does not have a preference for AES.  This should help to
-         identify why encrypting to several recipients falls back to
-         3DES. */
-      if (opt.verbose && cfx.dek->algo == CIPHER_ALGO_3DES)
-        warn_missing_aes_from_pklist (pk_list);
-    }
-  else
-    {
-      if (!opt.expert
-          && (select_algo_from_prefs (pk_list, PREFTYPE_SYM,
-                                      opt.def_cipher_algo, NULL)
-              != opt.def_cipher_algo))
-        {
-          log_info(_("WARNING: forcing symmetric cipher %s (%d)"
-                     " violates recipient preferences\n"),
-                   openpgp_cipher_algo_name (opt.def_cipher_algo),
-                   opt.def_cipher_algo);
-        }
-
-      cfx.dek->algo = opt.def_cipher_algo;
-    }
-
-  if (openpgp_cipher_blocklen (cfx.dek->algo) < 16
-      && !opt.flags.allow_old_cipher_algos)
-    {
-      log_error (_("cipher algorithm '%s' may not be used for encryption\n"),
-		 openpgp_cipher_algo_name (cfx.dek->algo));
-      if (!opt.quiet)
-        log_info (_("(use option \"%s\" to override)\n"),
-                  "--allow-old-cipher-algos");
-      rc = gpg_error (GPG_ERR_CIPHER_ALGO);
-      goto leave;
-    }
-
-  /* Check compliance.  */
-  if (! gnupg_cipher_is_allowed (opt.compliance, 1, cfx.dek->algo,
-                                 GCRY_CIPHER_MODE_CFB))
-    {
-      log_error (_("cipher algorithm '%s' may not be used in %s mode\n"),
-		 openpgp_cipher_algo_name (cfx.dek->algo),
-		 gnupg_compliance_option_string (opt.compliance));
-      rc = gpg_error (GPG_ERR_CIPHER_ALGO);
-      goto leave;
-    }
-
-  if (!gnupg_rng_is_compliant (opt.compliance))
-    {
-      rc = gpg_error (GPG_ERR_FORBIDDEN);
-      log_error (_("%s is not compliant with %s mode\n"),
-                 "RNG",
-                 gnupg_compliance_option_string (opt.compliance));
-      write_status_error ("random-compliance", rc);
-      goto leave;
-    }
-
-  compliant = gnupg_cipher_is_compliant (CO_DE_VS, cfx.dek->algo,
-                                         GCRY_CIPHER_MODE_CFB);
-
-  {
-    pk_list_t pkr;
-
-    for (pkr = pk_list; pkr; pkr = pkr->next)
-      {
-        PKT_public_key *pk = pkr->pk;
-        unsigned int nbits = nbits_from_pk (pk);
-
-        if (!gnupg_pk_is_compliant (opt.compliance, pk->pubkey_algo, 0,
-                                    pk->pkey, nbits, NULL))
-          log_info (_("WARNING: key %s is not suitable for encryption"
-                      " in %s mode\n"),
-                    keystr_from_pk (pk),
-                    gnupg_compliance_option_string (opt.compliance));
-
-        if (compliant
-            && !gnupg_pk_is_compliant (CO_DE_VS, pk->pubkey_algo, 0, pk->pkey,
-                                       nbits, NULL))
-          compliant = 0;
-      }
-
-  }
-
-  if (compliant)
-    write_status_strings (STATUS_ENCRYPTION_COMPLIANCE_MODE,
-                          gnupg_status_compliance_flag (CO_DE_VS),
-                          NULL);
+  rc = check_encryption_compliance (cfx.dek, pk_list);
+  if (rc)
+    goto leave;
 
   cfx.dek->use_aead = use_aead (pk_list, cfx.dek->algo);
   if (!cfx.dek->use_aead)
     cfx.dek->use_mdc = !!use_mdc (pk_list, cfx.dek->algo);
 
-  /* Only do the is-file-already-compressed check if we are using a
-   * MDC or AEAD.  This forces compressed files to be re-compressed if
-   * we do not have a MDC to give some protection against chosen
-   * ciphertext attacks. */
-  if (do_compress
-      && (cfx.dek->use_mdc || cfx.dek->use_aead)
-      && is_file_compressed (filename, &rc2))
-    {
-      if (opt.verbose)
-        log_info(_("'%s' already compressed\n"), filename);
-      do_compress = 0;
-    }
-  if (rc2)
-    {
-      rc = rc2;
-      goto leave;
-    }
-
   make_session_key (cfx.dek);
   if (DBG_CRYPTO)
     log_printhex (cfx.dek->key, cfx.dek->keylen, "DEK is: ");
@@ -898,11 +933,10 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   if (filename && *filename
       && !iobuf_is_pipe_filename (filename) && !opt.textmode )
     {
-      off_t tmpsize;
-      int overflow;
+      uint64_t tmpsize;
 
-      if ( !(tmpsize = iobuf_get_filelength(inp, &overflow))
-           && !overflow && opt.verbose)
+      tmpsize = iobuf_get_filelength (inp);
+      if (!tmpsize && opt.verbose)
         log_info(_("WARNING: '%s' is an empty file\n"), filename );
       /* We can't encode the length of very large files because
          OpenPGP uses only 32 bit for file sizes.  So if the size
@@ -916,6 +950,26 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   else
     filesize = opt.set_filesize ? opt.set_filesize : 0; /* stdin */
 
+  /* Register the cipher filter. */
+  iobuf_push_filter (out,
+                     cfx.dek->use_aead? cipher_filter_aead
+                     /**/             : cipher_filter_cfb,
+                     &cfx);
+
+  /* Only do the is-file-already-compressed check if we are using a
+   * MDC or AEAD.  This forces compressed files to be re-compressed if
+   * we do not have a MDC to give some protection against chosen
+   * ciphertext attacks. */
+  if (do_compress
+      && (cfx.dek->use_mdc || cfx.dek->use_aead)
+      && !opt.explicit_compress_option
+      && is_file_compressed (inp))
+    {
+      if (opt.verbose)
+        log_info(_("'%s' already compressed\n"), filename? filename: "[stdin]");
+      do_compress = 0;
+    }
+
   if (!opt.no_literal)
     {
       pt->timestamp = make_timestamp();
@@ -930,12 +984,6 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
   else
     cfx.datalen = filesize && !do_compress ? filesize : 0;
 
-  /* Register the cipher filter. */
-  iobuf_push_filter (out,
-                     cfx.dek->use_aead? cipher_filter_aead
-                     /**/             : cipher_filter_cfb,
-                     &cfx);
-
   /* Register the compress filter. */
   if (do_compress)
     {
@@ -977,19 +1025,14 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
     {
       /* User requested not to create a literal packet, so we copy the
          plain data. */
-      byte copy_buffer[4096];
-      int  bytes_copied;
-      while ((bytes_copied = iobuf_read (inp, copy_buffer, 4096)) != -1)
-        {
-          rc = iobuf_write (out, copy_buffer, bytes_copied);
-          if (rc)
-            {
-              log_error ("copying input to output failed: %s\n",
-                         gpg_strerror (rc));
-              break;
-            }
-        }
-      wipememory (copy_buffer, 4096); /* Burn the buffer. */
+      iobuf_copy (out, inp);
+      if ((rc = iobuf_error (inp)))
+        log_error (_("error reading '%s': %s\n"),
+                   iobuf_get_fname_nonnull (inp), gpg_strerror (rc));
+      else if ((rc = iobuf_error (out)))
+        log_error (_("error writing '%s': %s\n"),
+                   iobuf_get_fname_nonnull (out), gpg_strerror (rc));
+
     }
 
   /* Finish the stuff. */
@@ -1016,6 +1059,90 @@ encrypt_crypt (ctrl_t ctrl, int filefd, const char *filename,
 }
 
 
+/* Re-encrypt files with a set of new recipients.  Note that this
+ * function is called by decrypt_message.  INFP is the iobuf from the
+ * input file which is positioned right after the pubkey_enc and
+ * symkey_enc packets.  */
+gpg_error_t
+reencrypt_to_new_recipients (ctrl_t ctrl, int armor, const char *filename,
+                             iobuf_t infp, strlist_t recipients,
+                             DEK *dek, struct pubkey_enc_list *pkenc_list)
+{
+  gpg_error_t err;
+  int save_no_encrypt_to;
+  pk_list_t newpk_list = NULL;
+  iobuf_t outfp = NULL;
+  armor_filter_context_t *outafx = NULL;
+  PACKET pkt;
+  struct pubkey_enc_list *el;
+  unsigned int count;
+
+  /* Get the keys for all additional recipients but do not encrypt to
+   * the encrypt-to keys. */
+  save_no_encrypt_to = opt.no_encrypt_to;
+  opt.no_encrypt_to = 1;
+  err = build_pk_list (ctrl, recipients, &newpk_list);
+  opt.no_encrypt_to = save_no_encrypt_to;
+  if (err)
+    goto leave;
+
+  /* Note that we use by default the suffixes .gpg or .asc */
+  err = open_outfile (GNUPG_INVALID_FD, filename, armor? 1:0, 0, &outfp);
+  if (err)
+    goto leave;
+
+  if (armor)
+    {
+      outafx = new_armor_context ();
+      push_armor_filter (outafx, outfp);
+    }
+
+  /* Write the new recipients first.  */
+  err = write_pubkey_enc_from_list (ctrl, newpk_list, dek, outfp);
+  if (err)
+    goto leave;
+
+  /* The write the old recipients in --add-recipients mode.  */
+  for (count=0, el = pkenc_list; el; el = el->next, count++)
+    if (!ctrl->clear_recipients)
+      {
+        if (opt.verbose)
+          show_encrypted_for_user_info (ctrl, 0, &el->d, dek);
+        init_packet (&pkt);
+        pkt.pkttype = PKT_PUBKEY_ENC;
+        pkt.pkt.pubkey_enc = &el->d;
+        err = build_packet (outfp, &pkt);
+        if (err)
+          log_error ("build_packet(pubkey_enc) failed: %s\n",
+                     gpg_strerror (err));
+      }
+  if (ctrl->clear_recipients && opt.verbose)
+    log_info (_("number of removed recipients: %u\n"), count);
+
+  iobuf_put (outfp, ctrl->last_read_ctb);
+
+  /* Finally copy the bulk of the message.  */
+  iobuf_copy (outfp, infp);
+  if ((err = iobuf_error (infp)))
+    log_error (_("error reading '%s': %s\n"),
+               iobuf_get_fname_nonnull (infp), gpg_strerror (err));
+  else if ((err = iobuf_error (outfp)))
+    log_error (_("error writing '%s': %s\n"),
+               iobuf_get_fname_nonnull (outfp), gpg_strerror (err));
+
+
+ leave:
+  if (err)
+    iobuf_cancel (outfp);
+  else
+    iobuf_close (outfp);
+  release_armor_context (outafx);
+  release_pk_list (newpk_list);
+  return err;
+}
+
+
+
 /*
  * Filter to do a complete public key encryption.
  */
@@ -1035,41 +1162,13 @@ encrypt_filter (void *opaque, int control,
     {
       if ( !efx->header_okay )
         {
-          efx->cfx.dek = xmalloc_secure_clear ( sizeof *efx->cfx.dek );
-          if ( !opt.def_cipher_algo  )
-            {
-              /* Try to get it from the prefs. */
-              efx->cfx.dek->algo =
-                select_algo_from_prefs (efx->pk_list, PREFTYPE_SYM, -1, NULL);
-              if (efx->cfx.dek->algo == -1 )
-                {
-                  /* Because 3DES is implicitly in the prefs, this can
-                     only happen if we do not have any public keys in
-                     the list.  */
-                  efx->cfx.dek->algo = DEFAULT_CIPHER_ALGO;
-                }
+          efx->header_okay = 1;
 
-              /* In case 3DES has been selected, print a warning if
-                 any key does not have a preference for AES.  This
-                 should help to identify why encrypting to several
-                 recipients falls back to 3DES. */
-              if (opt.verbose
-                  && efx->cfx.dek->algo == CIPHER_ALGO_3DES)
-                warn_missing_aes_from_pklist (efx->pk_list);
-	    }
-          else
-            {
-	      if (!opt.expert
-                  && select_algo_from_prefs (efx->pk_list,PREFTYPE_SYM,
-                                             opt.def_cipher_algo,
-                                             NULL) != opt.def_cipher_algo)
-		log_info(_("forcing symmetric cipher %s (%d) "
-			   "violates recipient preferences\n"),
-			 openpgp_cipher_algo_name (opt.def_cipher_algo),
-			 opt.def_cipher_algo);
+          efx->cfx.dek = create_dek_with_warnings (efx->pk_list);
 
-	      efx->cfx.dek->algo = opt.def_cipher_algo;
-	    }
+          rc = check_encryption_compliance (efx->cfx.dek, efx->pk_list);
+          if (rc)
+            return rc;
 
           efx->cfx.dek->use_aead = use_aead (efx->pk_list, efx->cfx.dek->algo);
           if (!efx->cfx.dek->use_aead)
@@ -1097,7 +1196,6 @@ encrypt_filter (void *opaque, int control,
                              /**/                  : cipher_filter_cfb,
                              &efx->cfx);
 
-          efx->header_okay = 1;
         }
       rc = iobuf_write (a, buf, size);
 
@@ -1132,12 +1230,13 @@ write_pubkey_enc (ctrl_t ctrl,
   enc->pubkey_algo = pk->pubkey_algo;
   keyid_from_pk( pk, enc->keyid );
   enc->throw_keyid = throw_keyid;
+  enc->seskey_algo = dek->algo;  /* (Used only by PUBKEY_ALGO_KYBER.) */
 
   /* Okay, what's going on: We have the session key somewhere in
    * the structure DEK and want to encode this session key in an
    * integer value of n bits. pubkey_nbits gives us the number of
    * bits we have to use.  We then encode the session key in some
-   * way and we get it back in the big intger value FRAME.  Then
+   * way and we get it back in the big integer value FRAME.  Then
    * we use FRAME, the public key PK->PKEY and the algorithm
    * number PK->PUBKEY_ALGO and pass it to pubkey_encrypt which
    * returns the encrypted value in the array ENC->DATA.  This
@@ -1147,23 +1246,14 @@ write_pubkey_enc (ctrl_t ctrl,
    * build_packet().  */
   frame = encode_session_key (pk->pubkey_algo, dek,
                               pubkey_nbits (pk->pubkey_algo, pk->pkey));
-  rc = pk_encrypt (pk->pubkey_algo, enc->data, frame, pk, pk->pkey);
+  rc = pk_encrypt (pk, frame, dek->algo, enc->data);
   gcry_mpi_release (frame);
   if (rc)
     log_error ("pubkey_encrypt failed: %s\n", gpg_strerror (rc) );
   else
     {
       if ( opt.verbose )
-        {
-          char *ustr = get_user_id_string_native (ctrl, enc->keyid);
-          log_info (_("%s/%s.%s encrypted for: \"%s\"\n"),
-                    openpgp_pk_algo_name (enc->pubkey_algo),
-                    openpgp_cipher_algo_name (dek->algo),
-                    dek->use_aead? openpgp_aead_algo_name (dek->use_aead)
-                    /**/         : "CFB",
-                    ustr );
-          xfree (ustr);
-        }
+        show_encrypted_for_user_info (ctrl, pk->pubkey_usage, enc, dek);
       /* And write it. */
       init_packet (&pkt);
       pkt.pkttype = PKT_PUBKEY_ENC;
@@ -1229,7 +1319,8 @@ encrypt_crypt_files (ctrl_t ctrl, int nfiles, char **files, strlist_t remusr)
             }
           line[strlen(line)-1] = '\0';
           print_file_status(STATUS_FILE_START, line, 2);
-          rc = encrypt_crypt (ctrl, -1, line, remusr, 0, NULL, -1);
+          rc = encrypt_crypt (ctrl, GNUPG_INVALID_FD, line, remusr,
+                              0, NULL, GNUPG_INVALID_FD);
           if (rc)
             log_error ("encryption of '%s' failed: %s\n",
                        print_fname_stdin(line), gpg_strerror (rc) );
@@ -1241,7 +1332,8 @@ encrypt_crypt_files (ctrl_t ctrl, int nfiles, char **files, strlist_t remusr)
       while (nfiles--)
         {
           print_file_status(STATUS_FILE_START, *files, 2);
-          if ( (rc = encrypt_crypt (ctrl, -1, *files, remusr, 0, NULL, -1)) )
+          if ((rc = encrypt_crypt (ctrl, GNUPG_INVALID_FD, *files, remusr,
+                                   0, NULL, GNUPG_INVALID_FD)))
             log_error("encryption of '%s' failed: %s\n",
                       print_fname_stdin(*files), gpg_strerror (rc) );
           write_status( STATUS_FILE_DONE );
