@@ -43,7 +43,7 @@ static FILE *statusfp;
 /* Data used to assuciate an Assuan context with local server data */
 struct server_local_s {
   assuan_context_t assuan_ctx;
-  int message_fd;
+  estream_t message_fp;
   int list_internal;
   int list_external;
   int list_to_output;           /* Write keylistings to the output fd. */
@@ -130,12 +130,12 @@ data_line_cookie_close (void *cookie)
 
 
 static void
-close_message_fd (ctrl_t ctrl)
+close_message_fp (ctrl_t ctrl)
 {
-  if (ctrl->server_local->message_fd != -1)
+  if (ctrl->server_local->message_fp)
     {
-      close (ctrl->server_local->message_fd);
-      ctrl->server_local->message_fd = -1;
+      es_fclose (ctrl->server_local->message_fp);
+      ctrl->server_local->message_fp = NULL;
     }
 }
 
@@ -287,6 +287,17 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
           ctrl->offline = i;
         }
     }
+  else if (!strcmp (key, "always-trust"))
+    {
+      /* We ignore this option if gpgsm has been started with
+         --always-trust (which also sets offline) and if
+         --require-compliance is active */
+      if (!opt.always_trust && !opt.require_compliance)
+        {
+          int i = *value? !!atoi (value) : 1;
+          ctrl->always_trust = i;
+        }
+    }
   else if (!strcmp (key, "request-origin"))
     {
       if (!opt.request_origin)
@@ -320,7 +331,8 @@ reset_notify (assuan_context_t ctx, char *line)
   gpgsm_release_certlist (ctrl->server_local->signerlist);
   ctrl->server_local->recplist = NULL;
   ctrl->server_local->signerlist = NULL;
-  close_message_fd (ctrl);
+  ctrl->always_trust = 0;
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
   return 0;
@@ -451,20 +463,26 @@ cmd_encrypt (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   certlist_t cl;
-  int inp_fd, out_fd;
+  gnupg_fd_t inp_fd;
+  gnupg_fd_t out_fd;
+  estream_t inp_fp;
   estream_t out_fp;
   int rc;
 
   (void)line;
 
-  inp_fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
-  if (inp_fd == -1)
+  inp_fd = assuan_get_input_fd (ctx);
+  if (inp_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
-  out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
-  if (out_fd == -1)
+  out_fd = assuan_get_output_fd (ctx);
+  if (out_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
 
-  out_fp = es_fdopen_nc (out_fd, "w");
+  inp_fp = open_stream_nc (inp_fd, "r");
+  if (!inp_fp)
+    return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
+
+  out_fp = open_stream_nc (out_fd, "w");
   if (!out_fp)
     return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
 
@@ -483,13 +501,15 @@ cmd_encrypt (assuan_context_t ctx, char *line)
   if (!rc)
     rc = gpgsm_encrypt (assuan_get_pointer (ctx),
                         ctrl->server_local->recplist,
-                        inp_fd, out_fp);
+                        inp_fp, out_fp);
+  es_fclose (inp_fp);
   es_fclose (out_fp);
 
   gpgsm_release_certlist (ctrl->server_local->recplist);
   ctrl->server_local->recplist = NULL;
-  /* Close and reset the fd */
-  close_message_fd (ctrl);
+  ctrl->always_trust = 0;
+  /* Close and reset the fp and the fds */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
   return rc;
@@ -508,30 +528,37 @@ static gpg_error_t
 cmd_decrypt (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
-  int inp_fd, out_fd;
+  gnupg_fd_t inp_fd;
+  gnupg_fd_t out_fd;
+  estream_t inp_fp;
   estream_t out_fp;
   int rc;
 
   (void)line;
 
-  inp_fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
-  if (inp_fd == -1)
+  inp_fd = assuan_get_input_fd (ctx);
+  if (inp_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
-  out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
-  if (out_fd == -1)
+  out_fd = assuan_get_output_fd (ctx);
+  if (out_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
 
-  out_fp = es_fdopen_nc (out_fd, "w");
+  inp_fp = open_stream_nc (inp_fd, "r");
+  if (!inp_fp)
+    return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
+
+  out_fp = open_stream_nc (out_fd, "w");
   if (!out_fp)
     return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
 
   rc = start_audit_session (ctrl);
   if (!rc)
-    rc = gpgsm_decrypt (ctrl, inp_fd, out_fp);
+    rc = gpgsm_decrypt (ctrl, inp_fp, out_fp);
+  es_fclose (inp_fp);
   es_fclose (out_fp);
 
   /* Close and reset the fds. */
-  close_message_fd (ctrl);
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
 
@@ -553,30 +580,36 @@ cmd_verify (assuan_context_t ctx, char *line)
 {
   int rc;
   ctrl_t ctrl = assuan_get_pointer (ctx);
-  int fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
-  int out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
+  gnupg_fd_t fd = assuan_get_input_fd (ctx);
+  gnupg_fd_t out_fd = assuan_get_output_fd (ctx);
+  estream_t fp = NULL;
   estream_t out_fp = NULL;
 
   (void)line;
 
-  if (fd == -1)
+  if (fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
 
-  if (out_fd != -1)
+  fp = open_stream_nc (fd, "r");
+  if (!fp)
+    return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
+
+  if (out_fd != GNUPG_INVALID_FD)
     {
-      out_fp = es_fdopen_nc (out_fd, "w");
+      out_fp = open_stream_nc (out_fd, "w");
       if (!out_fp)
         return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
     }
 
   rc = start_audit_session (ctrl);
   if (!rc)
-    rc = gpgsm_verify (assuan_get_pointer (ctx), fd,
-                       ctrl->server_local->message_fd, out_fp);
+    rc = gpgsm_verify (assuan_get_pointer (ctx), fp,
+                       ctrl->server_local->message_fp, out_fp);
+  es_fclose (fp);
   es_fclose (out_fp);
 
-  /* Close and reset the fd.  */
-  close_message_fd (ctrl);
+  /* Close and reset the fp and the fd.  */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
 
@@ -594,32 +627,39 @@ static gpg_error_t
 cmd_sign (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
-  int inp_fd, out_fd;
+  gnupg_fd_t inp_fd;
+  gnupg_fd_t out_fd;
+  estream_t inp_fp;
   estream_t out_fp;
   int detached;
   int rc;
 
-  inp_fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
-  if (inp_fd == -1)
+  inp_fd = assuan_get_input_fd (ctx);
+  if (inp_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
-  out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
-  if (out_fd == -1)
+  out_fd = assuan_get_output_fd (ctx);
+  if (out_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
 
   detached = has_option (line, "--detached");
 
-  out_fp = es_fdopen_nc (out_fd, "w");
+  inp_fp = open_stream_nc (inp_fd, "r");
+  if (!inp_fp)
+    return set_error (gpg_err_code_from_syserror (), "fdopen() failed");
+
+  out_fp = open_stream_nc (out_fd, "w");
   if (!out_fp)
     return set_error (GPG_ERR_ASS_GENERAL, "fdopen() failed");
 
   rc = start_audit_session (ctrl);
   if (!rc)
     rc = gpgsm_sign (assuan_get_pointer (ctx), ctrl->server_local->signerlist,
-                     inp_fd, detached, out_fp);
+                     inp_fp, detached, out_fp);
+  es_fclose (inp_fp);
   es_fclose (out_fp);
 
-  /* close and reset the fd */
-  close_message_fd (ctrl);
+  /* close and reset the fp and the fds */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
 
@@ -644,18 +684,24 @@ cmd_import (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   int rc;
-  int fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
+  gnupg_fd_t fd = assuan_get_input_fd (ctx);
   int reimport = has_option (line, "--re-import");
+  estream_t fp;
 
   (void)line;
 
-  if (fd == -1)
+  if (fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
 
-  rc = gpgsm_import (assuan_get_pointer (ctx), fd, reimport);
+  fp = open_stream_nc (fd, "r");
+  if (!fp)
+    return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
 
-  /* close and reset the fd */
-  close_message_fd (ctrl);
+  rc = gpgsm_import (assuan_get_pointer (ctx), fp, reimport);
+  es_fclose (fp);
+
+  /* close and reset the fp and the fds */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
 
@@ -756,15 +802,15 @@ cmd_export (assuan_context_t ctx, char *line)
     }
   else
     {
-      int fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
+      gnupg_fd_t fd = assuan_get_output_fd (ctx);
       estream_t out_fp;
 
-      if (fd == -1)
+      if (fd == GNUPG_INVALID_FD)
         {
           free_strlist (list);
           return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
         }
-      out_fp = es_fdopen_nc (fd, "w");
+      out_fp = open_stream_nc (fd, "w");
       if (!out_fp)
         {
           free_strlist (list);
@@ -780,8 +826,8 @@ cmd_export (assuan_context_t ctx, char *line)
     }
 
   free_strlist (list);
-  /* Close and reset the fds. */
-  close_message_fd (ctrl);
+  /* Close and reset the fp and the fds. */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
   return 0;
@@ -829,8 +875,8 @@ cmd_delkeys (assuan_context_t ctx, char *line)
   rc = gpgsm_delete (ctrl, list);
   free_strlist (list);
 
-  /* close and reset the fd */
-  close_message_fd (ctrl);
+  /* close and reset the fp and the fds */
+  close_message_fp (ctrl);
   assuan_close_input_fd (ctx);
   assuan_close_output_fd (ctx);
 
@@ -864,19 +910,18 @@ static gpg_error_t
 cmd_message (assuan_context_t ctx, char *line)
 {
   int rc;
-  gnupg_fd_t sysfd;
-  int fd;
+  gnupg_fd_t fd;
+  estream_t fp;
   ctrl_t ctrl = assuan_get_pointer (ctx);
 
-  rc = assuan_command_parse_fd (ctx, line, &sysfd);
+  rc = assuan_command_parse_fd (ctx, line, &fd);
   if (rc)
     return rc;
 
-
-  fd = translate_sys2libc_fd (sysfd, 0);
-  if (fd == -1)
+  fp = open_stream_nc (fd, "r");
+  if (!fp)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
-  ctrl->server_local->message_fd = fd;
+  ctrl->server_local->message_fp = fp;
   return 0;
 }
 
@@ -1011,14 +1056,14 @@ do_listkeys (assuan_context_t ctx, char *line, int mode)
 
   if (ctrl->server_local->list_to_output)
     {
-      int outfd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
+      gnupg_fd_t outfd = assuan_get_output_fd (ctx);
 
-      if ( outfd == -1 )
+      if ( outfd == GNUPG_INVALID_FD )
         {
           free_strlist (list);
           return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
         }
-      fp = es_fdopen_nc (outfd, "w");
+      fp = open_stream_nc (outfd, "w");
       if (!fp)
         {
           free_strlist (list);
@@ -1087,24 +1132,24 @@ static gpg_error_t
 cmd_genkey (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
-  int inp_fd, out_fd;
+  gnupg_fd_t inp_fd, out_fd;
   estream_t in_stream, out_stream;
   int rc;
 
   (void)line;
 
-  inp_fd = translate_sys2libc_fd (assuan_get_input_fd (ctx), 0);
-  if (inp_fd == -1)
+  inp_fd = assuan_get_input_fd (ctx);
+  if (inp_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_INPUT, NULL);
-  out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
-  if (out_fd == -1)
+  out_fd = assuan_get_output_fd (ctx);
+  if (out_fd == GNUPG_INVALID_FD)
     return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
 
-  in_stream = es_fdopen_nc (inp_fd, "r");
+  in_stream = open_stream_nc (inp_fd, "r");
   if (!in_stream)
     return set_error (GPG_ERR_ASS_GENERAL, "es_fdopen failed");
 
-  out_stream = es_fdopen_nc (out_fd, "w");
+  out_stream = open_stream_nc (out_fd, "w");
   if (!out_stream)
     {
       es_fclose (in_stream);
@@ -1135,7 +1180,7 @@ static gpg_error_t
 cmd_getauditlog (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
-  int  out_fd;
+  gnupg_fd_t out_fd;
   estream_t out_stream;
   int opt_data, opt_html;
   int rc;
@@ -1156,11 +1201,11 @@ cmd_getauditlog (assuan_context_t ctx, char *line)
     }
   else
     {
-      out_fd = translate_sys2libc_fd (assuan_get_output_fd (ctx), 1);
-      if (out_fd == -1)
+      out_fd = assuan_get_output_fd (ctx);
+      if (out_fd == GNUPG_INVALID_FD)
         return set_error (GPG_ERR_ASS_NO_OUTPUT, NULL);
 
-      out_stream = es_fdopen_nc (out_fd, "w");
+      out_stream = open_stream_nc (out_fd, "w");
       if (!out_stream)
         {
           return set_error (GPG_ERR_ASS_GENERAL, "es_fdopen() failed");
@@ -1189,7 +1234,8 @@ static const char hlp_getinfo[] =
   "  agent-check - Return success if the agent is running.\n"
   "  cmd_has_option CMD OPT\n"
   "              - Returns OK if the command CMD implements the option OPT.\n"
-  "  offline     - Returns OK if the connection is in offline mode.";
+  "  offline     - Returns OK if the connection is in offline mode."
+  "  always-trust- Returns OK if the connection is in always-trust mode.";
 static gpg_error_t
 cmd_getinfo (assuan_context_t ctx, char *line)
 {
@@ -1248,6 +1294,11 @@ cmd_getinfo (assuan_context_t ctx, char *line)
     {
       rc = ctrl->offline? 0 : gpg_error (GPG_ERR_FALSE);
     }
+  else if (!strcmp (line, "always-trust"))
+    {
+      rc = (ctrl->always_trust || opt.always_trust)? 0
+           /**/                                    : gpg_error (GPG_ERR_FALSE);
+    }
   else
     rc = set_error (GPG_ERR_ASS_PARAMETER, "unknown value for WHAT");
 
@@ -1422,7 +1473,7 @@ gpgsm_server (certlist_t default_recplist)
   assuan_set_pointer (ctx, &ctrl);
   ctrl.server_local = xcalloc (1, sizeof *ctrl.server_local);
   ctrl.server_local->assuan_ctx = ctx;
-  ctrl.server_local->message_fd = -1;
+  ctrl.server_local->message_fp = NULL;
   ctrl.server_local->list_internal = 1;
   ctrl.server_local->list_external = 0;
   ctrl.server_local->default_recplist = default_recplist;
